﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=StaticCode
Version=7.28
@EndOfDesignText@
'Static code module

#Region Comment
	'This module uses code which has been filtered by "tidyCode"
	'The filtering makes a possible to do a fairly simple one-to-one translation
	'No external .js files are needed
	
	'1. "tidyCode" uses the dictionary (term.txt in the Objects folder) to substitute different spellings for language keywords
	'2. "restructure" maps a pattern of terms into another pattern of the same terms, it makes translation much simpler
	'3. "helper" adds functions and prototypes to help the JS engine interpret some B4X language elements
	
#End Region


Sub Process_Globals
End Sub

Sub JS(s As String) As String
	If s.StartsWith("end") Then
		s = "}"
	Else if s.startsWith("sub ") Then
		If s.IndexOf("(") = -1 Then s = s & " ()"
		s = restructure(s, "sub $0($1)", "function $0($1){")
	Else if s.startsWith("for ") Then
		If s.indexOf("each")>-1 Then
			s = restructure(s, "for each $0 in $1", "for (var $0 in $1){")
		Else If s.indexOf("step -") > -1 Then
			s = restructure(s, "for $0 = $1 to $2 step $3", "for (var $0 = $1; $0>$2-1; $0=$0$3){")
		Else If s.indexOf("step ") > -1 Then
			s = restructure(s, "for $0 = $1 to $2 step $3", "for (var $0 = $1; $0<$2+1; $0=$0+$3){")
		Else
			s = restructure(s, "for $0 = $1 to $2", "for (var $0 = $1; $0<$2+1; $0++){")
		End If
	Else if s.startsWith("if ") Then
		If s.indexOf("else")>-1 Then
			s = restructure(s, "if $0 then $1 else $2", "if ($0) {$1}else{$2}")
		Else If s.indexOf("then ")>-1 Then
			s = restructure(s, "if $0 then $1", "if ($0) {$1}")
		Else
			s = restructure(s, "if $0 then", "if ($0) {")
		End If
		Dim parts() As String = tidyCode.twoParts(s, "{")
		parts(0) = parts(0).Replace(" = ", " == ")
		s = parts(0) & "{" & parts(1)
	Else if s.startsWith("else if ") Then
		s = restructure(s, "else if $0 then", "}else if ($0){")
	Else if s.startsWith("do ") Then
		s = restructure(s, "do $0", "while ($0){")
	Else if s.startsWith("select case") Then
		s = restructure(s, "select case $0", "switch ($0){")
	Else if s.startsWith("case else") Then
		s = "default:"
	Else if s.startsWith("case") Then
		s = restructure(s, "case $0", "case $0:")
	Else if s.startsWith("else") Then
		s = "}else{"
	Else if s.startsWith("var ") Then
		If s.indexOf(":")>-1 Then s = restructure(s, "var $0: $0", "var $0 = []")
		Dim kx1 As Int = s.indexOf("(")
		Dim kx2 As Int = s.indexOf("=")
		If kx1>-1 And kx2 = -1 Then
			s = restructure(s, "var $0()", "var $0 = []")
			tidyCode.vectors.Put(s.substring(4), True)
		Else If kx1>-1 And kx2>kx1 Then
			s = restructure(s, "var $0()$1 = array$2($3)", "var $0 = [$3]")
			tidyCode.vectors.Put(s.substring2(4, s.IndexOf("=") - 1), True)
		End If
	End If

	'vectors require special treatment x(middleItem) => x[middleItem]
	For Each sk As String In tidyCode.vectors.Keys
		Do While s.indexOf(sk & "(")>-1
			s = restructure(s, "$0" & sk & "($1)$2", "$0 " & sk & "[$1] $2")
		Loop
	Next
	

	'Since the following handles elements that could appear multiple times in a line we need a loop				
	Do While s.indexOf("regex.split")>-1 
		s = restructure(s, "$0regex.split($1, $2)$3", "$0 $2.split ($1)$3")
	Loop
	Do While s.indexOf("regex.replaceAll")>-1
		s = restructure(s, "$0regex.replaceAll($1, $2, $3)$4", "$0 $2.replace(/$1/g, $3)$4")
	Loop
	Do While s.indexOf(".set")>-1
		s = restructure(s, "$0.set($1, $2)$3", "$0[$1] = $2$3")
	Loop
	Do While s.indexOf(".put")>-1
		s = restructure(s, "$0.put($1, $2)$3", "$0[$1] = $2$3")
	Loop
	Do While s.indexOf(".get")>-1
		s = restructure(s, "$0.get($1)$2", "$0[$1] $2")
	Loop
	Do While s.indexOf(".remove")>-1
		s = restructure(s, "$0.remove($1, $2)$3", "$0 $1.splice($2, $3-$2)$3")
	Loop
	Do While s.indexOf("asc(")>-1
		s = restructure(s, "$0asc($1)$2", "$0 $1.charCodeAt(0)$2")
	Loop
	Do While s.indexOf("array(")>-1
		s = restructure(s, "$0array($1)$2", "$0 [$1]$2")
	Loop
	Do While s.indexOf("containsKey(")>-1
		s = restructure(s, "$0containsKey($1)$2", "$0 [$1] != null$2")
	Loop

	
	'All expressions in JS end with a semicolon
	If (s.EndsWith("{") Or s.EndsWith("}") Or s.EndsWith(";")) = False Then
		If s.Length>0 Then s = s.trim & ";"
	End If
	
	'JS does not need 'As' in declarations
	Do While s.IndexOf(" as ") > -1
		Dim v() As String = tidyCode.twoParts(s, " as ")
		Dim k As Int = findWordEnd(v(1))
		If k>-1 Then 
			s = v(0) & v(1).SubString(k)		
		Else
			s = v(0)
		End If
	Loop
	
	'The next section cleans up some leftover language crumbs
	s = s.Replace(".keys", "")
	s = s.Replace("TAB", $""\t""$)
	s = s.Replace("CRLF",  $""\n""$)
	s = s.Replace(".~", "")
	s = s.Replace("^.", "")
	s = s.Replace(" & ", " + ")
	s = s.Replace("=[", "= [")

	If s.Trim.IndexOf(" ") = -1 And s.IndexOf("(") = -1  And s.IndexOf("}") = -1 Then s = s.Replace(";", "();")
	
	'Since there is no alert in a webview, myLog (see helper section below) returns a value that is logged in B4X
	s = s.Replace("alert(", "myLog(")
	Return s
End Sub

Sub restructure (s As String, olds As String, news As String) As String
	Dim v() As String = Regex.Split("\$", olds)
	Dim w() As String = Regex.Split("\$", news)
	
	Dim collect As List: collect.Initialize
	Dim lastK As Int = 0
	For j = 1 To v.Length-1
		v(j) = v(j).SubString(1)
		Dim k1 As Int = s.IndexOf2(v(j-1), lastK) + v(j-1).length
		Dim k2 As Int = s.length
		If v(j).Length > 0 Then k2 = s.IndexOf2(v(j), k1)
		If k1 >= s.Length Or k2<0 Then Exit
		Dim par As String = s.SubString2(k1, k2)
		If v(j) = ")" And par.EndsWith("(") Then  par = par & ")"
		collect.Add(par.trim)
		lastK = k2
	Next
	Dim sb As StringBuilder: sb.Initialize
	sb.Append(w(0))
	For j = 1 To w.Length-1
		Dim m As Int = w(j).SubString2(0, 1)
		If collect.Size > 0 And m<collect.Size Then
			Dim item As String = w(j).SubString(1)
			sb.Append(collect.Get(m) & item)
		End If
	Next
	Return sb.ToString
End Sub

Sub findWordEnd(s As String) As String
	For j = 0 To s.Length - 1
		Dim c As String = s.CharAt(j)
		If tidyCode.delimiters.IndexOf(c) > -1 Then Return j
	Next
	Return -1	
End Sub

Sub helper() As String
	'Add more as needed
	Dim result As String = _
		$"
			function myLog(t){return t;}
			function NumberFormat(f, x){return x;}
			String.prototype.replaceAll = function(search, replacement) {var target = this;return target.split(search).join(replacement);}
		"$

	result = result.Replace(CRLF & TAB & TAB & TAB, ";")
	Return result
End Sub

